//
//  ObjGHMM.m
//  GHMMCurveRecognition
//
//  Created by Marcelo Cicconet on 10/21/11.
//  Copyright 2011 Georgia Tech. All rights reserved.
//

#import "ObjGHMM.h"

@implementation ObjGHMM

@synthesize distanceToCurves, distanceToCurvesUsingAlpha, maxStates, likMaxStates;

- (id)init
{
    self = [super init];
    if (self) {
        // Initialization code here.
    }
    
    return self;
}

- (void)trainWithCurves:(NSMutableArray *)theCurves
{
    nCurves = (int)[theCurves count];
    arrayOfModels = [[NSMutableArray alloc] initWithCapacity:nCurves];
    
    nCurvePoints = (int *)malloc(nCurves*sizeof(int));
    
    maxStates = (int **)malloc(nCurves*sizeof(int *));
    likMaxStates = (float **)malloc(nCurves*sizeof(float *));;
    
    for (int curveIndex = 0; curveIndex < nCurves; curveIndex++) {
        Curve * curve = [theCurves objectAtIndex:curveIndex];
        nCurvePoints[curveIndex] = [curve numOfPoints];
        ModelGHMM * model = [[ModelGHMM alloc] initWithCurve:curve];
        [arrayOfModels addObject:model];
        [model release];
        
        maxStates[curveIndex] = (int *)malloc([curve numOfPoints]*sizeof(int));
        likMaxStates[curveIndex] = (float *)malloc([curve numOfPoints]*sizeof(float));
    }
    
    distanceToCurves = (float *)malloc(nCurves*sizeof(float));
    distanceToCurvesUsingAlpha = (float *)malloc(nCurves*sizeof(float));
}

- (void)testWithCurve:(Curve *)theCurve
{
    models = (ghmm_cmodel *)malloc(nCurves*sizeof(ghmm_cmodel));
    
    // load training
    int nTrain = (int)[arrayOfModels count];
    
    for (int trainIndex = 0; trainIndex < nTrain; trainIndex++) {
        ModelGHMM * model = [arrayOfModels objectAtIndex:trainIndex];
        
        models[trainIndex].model_type = GHMM_kContinuousHMM;
        models[trainIndex].N=model.n_states; /* states */
        models[trainIndex].M=1; /* density functions per state */
        models[trainIndex].dim=2; /* number of dimensions */
        models[trainIndex].cos=1; /* class of states */
        models[trainIndex].prior=-1; /* a priori probability */
        models[trainIndex].s = model.states;
        models[trainIndex].class_change = NULL;
        
//        if (ghmm_cmodel_check(&models[trainIndex]))
//        {
//            printf("ghmm_dmodel_check failed!\n");
//        }
//        fprintf(stdout,"model is ok\n");
        
        if (trainIndex == 0) {
            for (int i = 0; i < model.n_states; i++) {
                ghmm_cstate s = models[trainIndex].s[i];
                printf("state: %d\n", i);
                for (int j = 0; j < s.out_states; j++) {
                    printf("%f\t",s.out_a[0][j]);
                }
                printf("\n");
            }
        }
                
//        printf("type: %d\n", models[trainIndex].model_type);
//        printf("N: %d\n", models[trainIndex].N);
//        ghmm_cstate * s = models[trainIndex].s;
//        for (int i = 0; i < models[trainIndex].N; i++) {
//            printf("state: %d\n", i); 
//            ghmm_cstate s0 = s[i];
//            printf("in states: %d\n", s0.in_states);
//            printf("out states: %d\n", s0.out_states);
//            for (int i = 0; i < s0.in_states; i++) {
//                printf("%d\t", s0.in_id[i]);
//            }
//            printf("\n");
//            for (int i = 0; i < s0.out_states; i++) {
//                printf("%d\t", s0.out_id[i]);
//            }
//            printf("\n");
//        }
        
        int length = nCurvePoints[trainIndex]; // length of test curve
        int n_states = model.n_states;
        int dimension = 2;
        
        double ** alpha = (double **)malloc(length*sizeof(double *));
        for (int t = 0; t < length; t++) {
            alpha[t] = (double *)calloc(n_states,sizeof(double));
        }
    
        double ** betha = (double **)malloc(length*sizeof(double *));
        for (int t = 0; t < length; t++) {
            betha[t] = (double *)calloc(n_states,sizeof(double));
        }
        
        double ** nn_gamma = (double **)malloc(length*sizeof(double *)); // non-normalized gamma
        for (int t = 0; t < length; t++) {
            nn_gamma[t] = (double *)calloc(n_states,sizeof(double));
        }
        
        double * scale = (double *)malloc(length*sizeof(double));
        int T = length*dimension;
        
        double * O = (double *)malloc(length*dimension*sizeof(double));
        for (int t = 0; t < length; t++) {
            O[2*t] = theCurve.xPoints[t];
            O[2*t+1] = theCurve.yPoints[t];
        }
        
        double log_p;
        
        ghmm_cmodel_forward(&models[trainIndex], O, T, NULL, alpha, scale, &log_p);
        ghmm_cmodel_backward(&models[trainIndex], O, T, NULL, betha, scale);
        
        double * sum_gamma = (double *)malloc(length*sizeof(double));
        double distance = 0.0;
        double distance_alpha = 0.0;
        
//        double distanceToUnitySlope = 0.0;
//        int previousActualMapping;
        for (int t = 0; t < length; t++) {
            double sum = 0.0;
            for (int state = 0; state < n_states; state++) {            
                nn_gamma[t][state] = alpha[t][state]*betha[t][state];
                sum += nn_gamma[t][state];
            }
            sum_gamma[t] = sum;
            double max = 0.0;
            int ind_max;
            double max_alpha = 0.0;
            int ind_max_alpha;
//            if (trainIndex == 0) printf("test point: %d\n", t);
            for (int state = 0; state < n_states; state++) {            
                nn_gamma[t][state] /= sum;
                double value = nn_gamma[t][state];
                if (value > max) {
                    max = value;
                    ind_max = state;
                }
                
                value = alpha[t][state];
                if (value > max_alpha) {
                    max_alpha = value;
                    ind_max_alpha = state;
                }
//                if (trainIndex == 0) printf("%f\t", nn_gamma[t][state]);
            }
//            if (trainIndex == 0) printf("%d\t",ind_max);
            
            maxStates[trainIndex][t] = ind_max;
            likMaxStates[trainIndex][t] = max_alpha;
            
            double idealMapping = (double)t;
            double actualMapping = (double)ind_max;
            double portion = fabs(idealMapping-actualMapping); // range: [0, n_states-1]
            distance += portion;
            
            double actualMappingAlpha = (double)ind_max_alpha;
            double portion_alpha = fabs(idealMapping-actualMappingAlpha);
            distance_alpha += pow(portion_alpha, 1.0);

//            previousActualMapping = actualMapping;
        }
        
        distanceToCurves[trainIndex] = distance/(double)((length-1)*(n_states-1));
        distanceToCurvesUsingAlpha[trainIndex] = distance_alpha/(double)((length-1)*(n_states-1));
        
        for (int t = 0; t < length; t++) {
            free(alpha[t]);
            free(betha[t]);
            free(nn_gamma[t]);
        }
        free(O);
        free(alpha);
        free(betha);
        free(nn_gamma);
        free(scale);
        free(sum_gamma);
    }
    
    free(models);
}

- (void)dealloc {
    free(nCurvePoints);
    free(distanceToCurves);
    free(distanceToCurvesUsingAlpha);
    for (int i = 0; i < nCurves; i++) {
        free(maxStates[i]);
    }
    free(maxStates);
    for (int i = 0; i < nCurves; i++) {
        free(likMaxStates[i]);
    }
    free(likMaxStates);
    [arrayOfModels release];
    [super dealloc];
}

@end
